/*
   Copyright (c) 2009-2014, Jack Poulson
   All rights reserved.

   This file is part of Elemental and is under the BSD 2-Clause License, 
   which can be found in the LICENSE file in the root directory, or at 
   http://opensource.org/licenses/BSD-2-Clause
*/
#pragma once
#ifndef EL_PSEUDOSPECTRA_UTIL_REARRANGE_HPP
#define EL_PSEUDOSPECTRA_UTIL_REARRANGE_HPP

namespace El {
namespace pspec {

template<typename T>
inline void
ReshapeIntoGrid( Int realSize, Int imagSize, const Matrix<T>& x, Matrix<T>& X )
{
#if 0    
    X.Resize( imagSize, realSize );
    for( Int j=0; j<realSize; ++j )
    {
        auto XSub = X( IR(0,imagSize), IR(j,j+1) );
        auto xSub = x( IR(j*imagSize,(j+1)*imagSize), IR(0,1) );
        XSub = xSub;
    }
#else
    // The sequential case can be optimized much more heavily than in parallel
    X.Resize( imagSize, realSize, imagSize );
    MemCopy( X.Buffer(), x.LockedBuffer(), realSize*imagSize );
#endif
}

template<typename T>
inline void
ReshapeIntoGrid
( Int realSize, Int imagSize, 
  const AbstractDistMatrix<T>& x, AbstractDistMatrix<T>& X )
{
    X.SetGrid( x.Grid() );
    X.Resize( imagSize, realSize );

    auto xSub = std::unique_ptr<AbstractDistMatrix<T>>
    ( x.Construct(x.Grid(),x.Root()) );
    auto XSub = std::unique_ptr<AbstractDistMatrix<T>>
    ( X.Construct(X.Grid(),X.Root()) );

    for( Int j=0; j<realSize; ++j )
    {
              View( *XSub, X, IR(0,imagSize),                IR(j,j+1) );
        LockedView( *xSub, x, IR(j*imagSize,(j+1)*imagSize), IR(0,1)   );
        Copy( *xSub, *XSub );
    }
}

template<typename T>
inline void
RestoreOrdering
( const Matrix<Int>& preimage, Matrix<T>& x )
{
    DEBUG_ONLY(CallStackEntry cse("pspec::RestoreOrdering"))
    auto xCopy = x;
    const Int numShifts = preimage.Height();
    for( Int j=0; j<numShifts; ++j )
    {
        const Int dest = preimage.Get(j,0);
        x.Set( dest, 0, xCopy.Get(j,0) );
    }
}

template<typename T1,typename T2>
inline void
RestoreOrdering( const Matrix<Int>& preimage, Matrix<T1>& x, Matrix<T2>& y )
{
    DEBUG_ONLY(CallStackEntry cse("pspec::RestoreOrdering"))
    auto xCopy = x;
    auto yCopy = y;
    const Int numShifts = preimage.Height();
    for( Int j=0; j<numShifts; ++j )
    {
        const Int dest = preimage.Get(j,0);
        x.Set( dest, 0, xCopy.Get(j,0) );
        y.Set( dest, 0, yCopy.Get(j,0) );
    }
}

template<typename T>
inline void
RestoreOrdering
( const AbstractDistMatrix<Int>& preimage,
        AbstractDistMatrix<T>& x )
{
    DEBUG_ONLY(CallStackEntry cse("pspec::RestoreOrdering"))
    DistMatrix<Int,STAR,STAR> preimageCopy( preimage );
    DistMatrix<T,STAR,STAR> xCopy( x );
    const Int numShifts = preimage.Height();
    for( Int j=0; j<numShifts; ++j )
    {
        const Int dest = preimageCopy.Get(j,0);
        x.Set( dest, 0, xCopy.Get(j,0) );
    }
}

template<typename T1,typename T2>
inline void
RestoreOrdering
( const AbstractDistMatrix<Int>& preimage,
        AbstractDistMatrix<T1>& x,
        AbstractDistMatrix<T2>& y )
{
    DEBUG_ONLY(CallStackEntry cse("pspec::RestoreOrdering"))
    DistMatrix<Int,STAR,STAR> preimageCopy( preimage );
    DistMatrix<T1, STAR,STAR> xCopy( x );
    DistMatrix<T2, STAR,STAR> yCopy( y );
    const Int numShifts = preimage.Height();
    for( Int j=0; j<numShifts; ++j )
    {
        const Int dest = preimageCopy.Get(j,0);
        x.Set( dest, 0, xCopy.Get(j,0) );
        y.Set( dest, 0, yCopy.Get(j,0) );
    }
}

template<typename T1,typename T2>
inline void
ExtractList
( const std::vector<Matrix<T1>>& vecList, Matrix<T2>& list, Int i )
{
    DEBUG_ONLY(
        CallStackEntry cse("pspec::ExtractList");
        if( vecList.size() != 0 && vecList[0].Height() <= i )
            LogicError("Invalid index");
    )
    const Int numVecs = vecList.size();
    list.Resize( numVecs, 1 );
    for( Int k=0; k<numVecs; ++k )
        list.Set( k, 0, vecList[k].Get(i,0) );
}

template<typename T1,typename T2>
inline void
ExtractList
( const std::vector<Matrix<T1>>& matList, Matrix<T2>& list, Int i, Int j )
{
    DEBUG_ONLY(CallStackEntry cse("pspec::ExtractList"))
    const Int numMats = matList.size();
    list.Resize( numMats, 1 );
    for( Int k=0; k<numMats; ++k )
        list.Set( k, 0, matList[k].Get( i, j ) );
}

template<typename T1,typename T2>
inline void
PlaceList
( std::vector<Matrix<T1>>& vecList, const Matrix<T2>& list, Int i )
{
    DEBUG_ONLY(
        CallStackEntry cse("pspec::PlaceList");
        if( vecList.size() != 0 && vecList[0].Height() <= i )
            LogicError("Invalid index");
        if( vecList.size() != list.Height() )
            LogicError("List sizes do not match");
        if( list.Width() != 1 )
            LogicError("list should be a column vector");
    )
    const Int numVecs = vecList.size();
    for( Int k=0; k<numVecs; ++k )
        vecList[k].Set( i, 0, list.Get(k,0) );
}

template<typename T1,typename T2>
inline void
PlaceList
( std::vector<Matrix<T1>>& matList, const Matrix<T2>& list, Int i, Int j )
{
    DEBUG_ONLY(
        CallStackEntry cse("pspec::PlaceList");
        if( matList.size() != list.Height() )
            LogicError("List sizes do not match");
        if( list.Width() != 1 )
            LogicError("List assumed to be a column vector");
    )
    const Int numMats = matList.size();
    for( Int k=0; k<numMats; ++k )
        matList[k].Set( i, j, list.Get(k,0) );
}

template<typename T1,typename T2>
inline void
UpdateList
( std::vector<Matrix<T1>>& matList, const Matrix<T2>& list, Int i, Int j )
{
    DEBUG_ONLY(
        CallStackEntry cse("pspec::UpdateList");
        if( matList.size() != list.Height() )
            LogicError("List sizes do not match");
        if( list.Width() != 1 )
            LogicError("list assumed to be a column vector");
    )
    const Int numMats = matList.size();
    for( Int k=0; k<numMats; ++k )
        matList[k].Update( i, j, list.Get(k,0) );
}

template<typename T1,typename T2>
inline void
PushBackList
( std::vector<Matrix<T1>>& vecList, const Matrix<T2>& list )
{
    DEBUG_ONLY(
        CallStackEntry cse("pspec::PushBackList"); 
        if( vecList.size() != list.Height() )
            LogicError("List sizes do not match");
        if( list.Width() != 1 )
            LogicError("list assumed to be a column vector");
    )
    const Int numVecs = vecList.size();
    for( Int k=0; k<numVecs; ++k )
    {
        const Int m = vecList[k].Height();
        if( vecList[k].LDim() == m )
        {
            std::cerr << "Warning: reallocation required in PushBackList" 
                      << std::endl;
            auto A = vecList[k];
            vecList[k].Resize( m+1, 1 );
            for( Int i=0; i<m; ++i )
                vecList[k].Set( i, 0, A.Get(i,0) );
        }
        else
        {
            vecList[k].Resize( m+1, 1 );
        }
        vecList[k].Set( m, 0, list.Get(k,0) );
    }
}

} // namespace pspec
} // namespace El

#endif // ifndef EL_PSEUDOSPECTRA_UTIL_REARRANGE_HPP
