/*
   Copyright (c) 2009-2014, Jack Poulson
   All rights reserved.

   This file is part of Elemental and is under the BSD 2-Clause License, 
   which can be found in the LICENSE file in the root directory, or at 
   http://opensource.org/licenses/BSD-2-Clause
*/
#include "El.hpp"

namespace El {

template<typename T>
void ThreeValued( Matrix<T>& A, Int m, Int n, double p )
{
    DEBUG_ONLY(CallStackEntry cse("ThreeValued"))
    A.Resize( m, n );
    auto tripleCoin = [=]() -> T
    { 
        const double alpha = SampleUniform<double>(0,1);
        if( alpha <= p/2 ) return T(-1);
        else if( alpha <= p ) return T(1);
        else return T(0);
    };
    EntrywiseFill( A, std::function<T()>(tripleCoin) );
}

template<typename T>
void ThreeValued( AbstractDistMatrix<T>& A, Int m, Int n, double p )
{
    DEBUG_ONLY(CallStackEntry cse("ThreeValued"))
    A.Resize( m, n );
    if( A.RedundantRank() == 0 )
        ThreeValued( A.Matrix(), A.LocalHeight(), A.LocalWidth(), p );
    A.BroadcastOver( A.RedundantComm(), 0 );
}

template<typename T>
void ThreeValued( AbstractBlockDistMatrix<T>& A, Int m, Int n, double p )
{
    DEBUG_ONLY(CallStackEntry cse("ThreeValued"))
    A.Resize( m, n );
    if( A.RedundantRank() == 0 )
        ThreeValued( A.Matrix(), A.LocalHeight(), A.LocalWidth(), p );
    A.BroadcastOver( A.RedundantComm(), 0 );
}


#define PROTO(T) \
  template void ThreeValued \
  ( Matrix<T>& A, Int m, Int n, double p ); \
  template void ThreeValued \
  ( AbstractDistMatrix<T>& A, Int m, Int n, double p ); \
  template void ThreeValued \
  ( AbstractBlockDistMatrix<T>& A, Int m, Int n, double p );

#include "El/macros/Instantiate.h"

} // namespace El
